__version__ = "0.2.33"
__license__ = "MIT License"
__website__ = "https://code.exrny.com/opensource/vulcan-aws/"
__download_url__ = ('https://github.com/exrny/vulcan-aws/archive/'
                    '{}.tar.gz'.format(__version__))
